package common.world;

import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.function.Predicate;

import common.biome.Biome;
import common.block.Block;
import common.block.ITileEntityProvider;
import common.collect.Maps;
import common.entity.Entity;
import common.init.BlockRegistry;
import common.init.Blocks;
import common.log.Log;
import common.material.Material;
import common.rng.Random;
import common.tileentity.TileEntity;
import common.util.BlockPos;
import common.util.BoundingBox;
import common.util.ChunkPos;
import common.util.ClassInheritanceMultiMap;
import common.util.ExtMath;
import common.util.Facing;
import common.util.NibbleArray;
import common.worldgen.BiomeGenerator;
import common.worldgen.DebugStates;

public class Chunk {
	public final int xPos;
	public final int yPos;
	public final int zPos;
	private final World world;
	private BlockArray blocks = null;
	private final Map<BlockPos, TileEntity> tiles = Maps.<BlockPos, TileEntity>newHashMap();
	private final ClassInheritanceMultiMap<Entity> entities = new ClassInheritanceMultiMap(Entity.class);
	private final ConcurrentLinkedQueue<BlockPos> tileQueue = new ConcurrentLinkedQueue<BlockPos>();

	private boolean loaded;
	private boolean populated;
	private boolean updated;
	private boolean modified;
	private boolean hasEntity;
	private int lightChecks = 256;
	private long lastSave;
	private long inhabited;

	public Chunk(World world, int x, int y, int z) {
		this.world = world;
		this.xPos = x;
		this.yPos = y;
		this.zPos = z;
	}

	public Chunk(World world, short[] data, int x, int y, int z) {
		this(world, x, y, z);
		for(int bx = 0; bx < 16; ++bx) {
			for(int bz = 0; bz < 16; ++bz) {
				for(int by = 0; by < 16; ++by) {
					State state = BlockRegistry.STATEMAP.getByValue(data[bx << 4 | bz | by << 8]);
					if(state != null && state.getBlock().getMaterial() != Material.air) {
						if(this.blocks == null)
							this.blocks = new BlockArray();
						this.blocks.set(bx, by, bz, state);
					}
				}
			}
		}
	}
	
	public BlockArray getStorage() {
		return this.blocks;
	}
	
	public int getOpacity(BlockPos pos) {
		return this.getBlock(pos).getLightOpacity();
	}

	private int getOpacity(int x, int y, int z) {
		return this.getBlock0(x, y, z).getLightOpacity();
	}

	private Block getBlock0(int x, int y, int z) {
		Block block = Blocks.air;

		if(this.blocks != null) {
			block = this.blocks.getBlock(x, y, z);
		}

		return block;
	}

	public Block getBlock(int x, int y, int z) {
		return this.getBlock0(x & 15, y & 15, z & 15);
	}

	public Block getBlock(BlockPos pos) {
		return this.getBlock0(pos.getX() & 15, pos.getY() & 15, pos.getZ() & 15);
	}

	public State getState(BlockPos pos) {
		if(this.world.debug) {
			State state = null;
			
			if(pos.getY() == 1) {
				state = DebugStates.getState(pos.getX(), pos.getZ());
			}

			return state == null ? Blocks.air.getState() : state;
		}
		else {
			if(this.blocks != null) {
				int x = pos.getX() & 15;
				int y = pos.getY() & 15;
				int z = pos.getZ() & 15;
				return this.blocks.get(x, y, z);
			}

			return Blocks.air.getState();
		}
	}

	private int getMeta(int x, int y, int z) {
		if(this.blocks == null)
			return 0;
		return this.blocks.getMeta(x, y, z);
	}

	public int getMeta(BlockPos pos) {
		return this.getMeta(pos.getX() & 15, pos.getY() & 15, pos.getZ() & 15);
	}

	public State setState(BlockPos pos, State state) {
		int x = pos.getX() & 15;
		int y = pos.getY() & 15;
		int z = pos.getZ() & 15;

		State old = this.getState(pos);

		if(old == state) {
			return null;
		}
		else {
			Block block = state.getBlock();
			Block oldb = old.getBlock();

			if(this.blocks == null) {
				if(block == Blocks.air) {
					return null;
				}

				this.blocks = new BlockArray();
			}

			this.blocks.set(x, y & 15, z, state);

			if(oldb != block) {
				if(!this.world.client) {
					oldb.onBlockRemoved((AWorldServer)this.world, pos, old);
				}
				else if(oldb instanceof ITileEntityProvider) {
					this.world.removeTileEntity(pos);
				}
			}

			if(this.blocks.getBlock(x, y, z) != block) {
				return null;
			}
			else {
				if(oldb instanceof ITileEntityProvider) {
					TileEntity tile = this.getTileEntity(pos, TileEntity.EnumCreateEntityType.CHECK);

					if(tile != null) {
						tile.updateContainingBlockInfo();
					}
				}

				if(!this.world.client && oldb != block) {
					block.onBlockAdded((AWorldServer)this.world, pos, state);
				}

				if(block instanceof ITileEntityProvider) {
					TileEntity tile = this.getTileEntity(pos, TileEntity.EnumCreateEntityType.CHECK);

					if(tile == null) {
						tile = ((ITileEntityProvider)block).createNewTileEntity(this.world, block.getMetaFromState(state));
						this.world.setTileEntity(pos, tile);
					}

					if(tile != null) {
						tile.updateContainingBlockInfo();
					}
				}

				this.modified = true;
				return old;
			}
		}
	}

	public int getLight(BlockPos pos) {
		int x = pos.getX() & 15;
		int y = pos.getY() & 15;
		int z = pos.getZ() & 15;
		return this.blocks == null ? 0 : this.blocks.getLight(x, y, z);
	}

	public void setLight(BlockPos pos, int value) {
		int x = pos.getX() & 15;
		int y = pos.getY() & 15;
		int z = pos.getZ() & 15;

		if(this.blocks == null)
			this.blocks = new BlockArray();
		
		this.modified = true;

		this.blocks.setLight(x, y, z, value);
	}

	public int getLightSub(BlockPos pos, int amount) {
		int x = pos.getX() & 15;
		int y = pos.getY() & 15;
		int z = pos.getZ() & 15;

		if(this.blocks == null) {
			return !this.world.dimension.hasNoLight() && amount < 15 
					? 15 - amount : 0;
		}
		else {
			int l = this.world.dimension.hasNoLight() ? 0 : 15;
			l = l - amount;
			int b = this.blocks.getLight(x, y, z);

			if(b > l) {
				l = b;
			}

			return l;
		}
	}

	public void addEntity(Entity entity) {
		this.hasEntity = true;
		int x = ExtMath.floord(entity.posX / 16.0D);
		int y = ExtMath.floord(entity.posY / 16.0D);
		int z = ExtMath.floord(entity.posZ / 16.0D);

		if(x != this.xPos || y != this.yPos || z != this.zPos) {
			Log.JNI.warn("Falsche Position! (" + x + ", " + y + ", " + z + ") sollte (" + this.xPos + ", " + this.yPos + ", " + this.zPos + ") sein, " + entity);
			entity.setDead();
		}
		
		entity.addedToChunk = true;
		entity.chunkCoordX = this.xPos;
		entity.chunkCoordY = this.yPos;
		entity.chunkCoordZ = this.zPos;
		this.entities.add(entity);
	}

	public void removeEntity(Entity entity) {
		this.entities.remove(entity);
	}
	
	private TileEntity createNewTileEntity(BlockPos pos) {
		Block block = this.getBlock(pos);
		return !block.hasTileEntity() ? null : ((ITileEntityProvider)block).createNewTileEntity(this.world, this.getMeta(pos));
	}

	public TileEntity getTileEntity(BlockPos pos, TileEntity.EnumCreateEntityType type) {
		TileEntity tile = this.tiles.get(pos);

		if(tile == null) {
			if(type == TileEntity.EnumCreateEntityType.IMMEDIATE) {
				tile = this.createNewTileEntity(pos);
				this.world.setTileEntity(pos, tile);
			}
			else if(type == TileEntity.EnumCreateEntityType.QUEUED) {
				this.tileQueue.add(pos);
			}
		}
		else if(tile.isInvalid()) {
			this.tiles.remove(pos);
			return null;
		}

		return tile;
	}

	public void addTileEntity(TileEntity tile) {
		this.addTileEntity(tile.getPos(), tile);

		if(this.loaded) {
			this.world.addTileEntity(tile);
		}
	}

	public void addTileEntity(BlockPos pos, TileEntity tile) {
		tile.setWorldObj(this.world);
		tile.setPos(pos);

		if(this.getBlock(pos) instanceof ITileEntityProvider) {
			if(this.tiles.containsKey(pos)) {
				this.tiles.get(pos).invalidate();
			}

			if(tile.validate()) {
				this.setModified();
			}
			this.tiles.put(pos, tile);
		}
	}

	public void removeTileEntity(BlockPos pos) {
		if(this.loaded) {
			TileEntity tile = this.tiles.remove(pos);

			if(tile != null) {
				tile.invalidate();
			}
		}
	}

	public void onChunkLoad() {
		this.loaded = true;
		this.world.addTileEntities(this.tiles.values());

		for(Entity entity : this.entities) {
			entity.onChunkLoad();
		}

		this.world.loadEntities(this.entities);
	}

	public void onChunkUnload() {
		this.loaded = false;

		for(TileEntity tile : this.tiles.values()) {
			this.world.markTileEntityForRemoval(tile);
		}

		this.world.unloadEntities(this.entities);
	}

	public void setModified() {
		this.modified = true;
	}

	public void getEntities(Entity exclude, BoundingBox bb, List<Entity> list, Predicate<? super Entity> pred) {
		if(!this.entities.isEmpty()) {
			for(Entity entity : this.entities) {
				if(entity.getEntityBoundingBox().intersectsWith(bb) && entity != exclude) {
					if(pred == null || pred.test(entity)) {
						list.add(entity);
					}

					Entity[] parts = entity.getParts();

					if(parts != null) {
						for(int l = 0; l < parts.length; ++l) {
							entity = parts[l];

							if(entity != exclude && entity.getEntityBoundingBox().intersectsWith(bb) 
									&& (pred == null || pred.test(entity))) {
								list.add(entity);
							}
						}
					}
				}
			}
		}
	}

	public <T extends Entity> void getEntities(Class<? extends T> clazz, BoundingBox bb, List<T> list, Predicate<? super T> pred) {
		for(T entity : this.entities.getByClass(clazz)) {
			if(entity.getEntityBoundingBox().intersectsWith(bb) && (pred == null || pred.test(entity))) {
				list.add(entity);
			}
		}
	}

	public boolean isDirty(long time) {
		if(this.hasEntity && time != this.lastSave || this.modified)
			return true;
		return this.modified;
	}

	public boolean isDummy() {
		return false;
	}
	
	public void update() {
		this.updated = true;
		
		while(!this.tileQueue.isEmpty()) {
			BlockPos pos = (BlockPos)this.tileQueue.poll();

			if(this.getTileEntity(pos, TileEntity.EnumCreateEntityType.CHECK) == null && this.getBlock(pos).hasTileEntity()) {
				TileEntity tile = this.createNewTileEntity(pos);
				this.world.setTileEntity(pos, tile);
				this.world.markBlockRangeForRenderUpdate(pos, pos);
			}
		}
	}

	public boolean isPopulated() {
		return this.updated && this.populated;
	}

	public ChunkPos getPos() {
		return new ChunkPos(this.xPos, this.yPos, this.zPos);
	}

	public boolean isEmpty() {
		return this.blocks == null || this.blocks.isEmpty();
	}

	public void setStorage(BlockArray array) {
		this.blocks = array;
	}

	public void setData(byte[] data, boolean update, boolean clear) {
		int pos = 0;
		
		if(update) {
			if(this.blocks == null) {
				this.blocks = new BlockArray();
			}

			char[] blocks = this.blocks.getData();

			for(int k = 0; k < blocks.length; ++k) {
				blocks[k] = (char)((data[pos + 1] & 255) << 8 | data[pos] & 255);
				pos += 2;
			}
		}
		else if(clear && this.blocks != null) {
			this.blocks = null;
		}

		if(update && this.blocks != null) {
			NibbleArray light = this.blocks.getBlocklight();
			System.arraycopy(data, pos, light.getData(), 0, light.getData().length);
			pos += light.getData().length;
		}
		
		if(this.blocks != null && update) {
			this.blocks.update();
		}

		this.populated = true;

		for(TileEntity tile : this.tiles.values()) {
			tile.updateContainingBlockInfo();
		}
	}

	public void resetRelight() {
		this.lightChecks = 0;
	}

	/**
	 * Called once-per-chunk-per-tick, and advances the round-robin relight check
	 * index by up to 8 blocks at a time. In a worst-case scenario, can potentially
	 * take up to 1.6 seconds, calculated via (256/8)/20, to re-check all blocks
	 * in a chunk, which may explain lagging light updates on initial world
	 * generation.
	 */
	public void enqueueRelight() {
		BlockPos pos = new BlockPos(this.xPos << 4, 0, this.zPos << 4);

		for(int n = 0; n < 8; ++n) {
			if(this.lightChecks >= 256)
				return;

			int x = this.lightChecks % 16;
			int z = this.lightChecks / 16;
			++this.lightChecks;

			for(int y = 0; y < 16; ++y) {
				BlockPos block = pos.add(x, y, z);
				boolean edge = y == 0 || y == 15 || x == 0 || x == 15 || z == 0 || z == 15;

				if(this.blocks == null && edge || this.blocks != null 
						&& this.blocks.getBlock(x, y, z).getMaterial() == Material.air) {
					for(Facing face : Facing.values()) {
						BlockPos side = block.offset(face);

						if(this.world.getState(side).getBlock().getLightValue() > 0) {
							this.world.checkLight(side);
						}
					}

					this.world.checkLight(block);
				}
			}
		}
	}
	
	public boolean isLoaded() {
		return this.loaded;
	}

	public void setLoaded(boolean loaded) {
		this.loaded = loaded;
	}

	public World getWorld() {
		return this.world;
	}

	public Map<BlockPos, TileEntity> getTiles() {
		return this.tiles;
	}

	public ClassInheritanceMultiMap<Entity> getEntities() {
		return this.entities;
	}

	public boolean isTerrainPopulated() {
		return this.populated;
	}

	public void setTerrainPopulated(boolean populated) {
		this.populated = populated;
	}

	public void setModified(boolean modified) {
		this.modified = modified;
	}

	public void setHasEntities(boolean entities) {
		this.hasEntity = entities;
	}

	public void setSaved(long time) {
		this.lastSave = time;
	}

	public long getInhabited() {
		return this.inhabited;
	}

	public void setInhabited(long time) {
		this.inhabited = time;
	}
}
